const watcher = require('chokidar');
const fs = require('fs-extra');
const archiver = require('archiver');
const rimraf = require('rimraf');
const exec = require('child_process').exec;
const Path = require('path');
const colors = require('colors');
const TypesParser = require('./typesParser');

const buildPath = 'build/';
const distPath = 'dist/';
const srcPath = 'src/';
const blueprintPath = 'blueprint/';

class Builder {
  constructor() {
    this.buildQueue = {};
  }

  static parseTypes(group, resource, verbose = true) {
    const parser = new TypesParser(`${srcPath}${group}/${resource}/`, resource);
    const parsedCount = parser.parse();
    if (verbose) {
      console.log(`Parsed ${parsedCount} types for ${resource}`);
    }
  }

  static buildMany(resources, progressTick = null) {
    const promises = [];
    const failed = [];
    const succeed = [];
    resources.forEach((resourceData) => {
      promises.push(() => {
        return new Promise((resolve) => {
          Builder.buildExercise(resourceData.group, resourceData.resource, false).then(() => {
            succeed.push(resourceData);
            progressTick();
            resolve();
          }).catch(() => {
            failed.push(resourceData);
            progressTick();
            resolve();
          });
        });
      });
    });

    promises.reduce(function(cur, next) {
      return cur.then(next);
    }, Promise.resolve()).then(function() {
      if (succeed.length > 0 ) {
        console.log(`Successfully built ${succeed.length} resources:`.green);
        succeed.forEach((succ) => {
          console.log(`  - ${succ.group}/${succ.resource}`.green);
        });
      }
      if (failed.length > 0) {
        console.log('Build failed for:'.red);
        failed.forEach((fail) => {
          console.log(`  - ${fail.group}/${fail.resource}`.red);
        });
      }
    });
  }

  static packExercise(group, resource) {
    const buildDir = `${buildPath}${group}/${resource}/`;
    const distDir = `${distPath}${group}/`;
    const distFile = `${distDir}${resource}.zip`;

    fs.ensureDirSync(distDir);
    rimraf.sync(distFile);

    return new Promise((resolve, reject) => {
      const archive = archiver('zip', {
        store: true
      });
      const output = fs.createWriteStream(distFile);

      output.on('close', () => {
        resolve();
      });
      archive.pipe(output);
      archive.directory(buildDir, '/');
      archive.finalize();
    });
  }

  static buildExercise(group, resource, verbose = true) {
    return new Promise((resolve, reject) => {
      this.compileExercise(group, resource, verbose).then(() => {
        const sourceDir = `${srcPath}${group}/${resource}/`;
        const buildDir = `${buildPath}${group}/${resource}/`;

        fs.ensureDirSync(`${buildPath}${group}/`);
        fs.ensureDirSync(buildDir);

        rimraf.sync(buildDir);
        fs.copySync(sourceDir, buildDir);
        rimraf.sync(`${buildDir}*.fla`);
        rimraf.sync(`${buildDir}*.ts`);
        rimraf.sync(`${buildDir}${resource}.html`);

        Builder.packExercise(group, resource).then(() => {
          resolve();
        }).catch(() => {
          reject();
        });
      }).catch(() => {
        reject();
      });
    });
  }

  static compileExercise(group, resource, verbose = true) {
    return new Promise((resolve, reject) => {
      Builder.generate(group, resource, false, verbose);
      const resourcePath = `${srcPath}${group}/${resource}/`;
      fs.stat(`${resourcePath}/main.ts`, (err) => {
        if (err) {
          if (verbose) {
            console.log(`Skipping ${group}/${resource} ...`.grey);
          }
          reject();
        }
        const ps = exec(`tsc --module amd --outFile ${resourcePath}/boot.js --lib DOM,ES5,ScriptHost,ES2015.Iterable --rootDir ${srcPath} --allowJs --target ES5 ${resourcePath}/boot.ts`);
        ps.stdout.pipe(process.stdout);
        ps.stderr.pipe(process.stderr);
        ps.on('exit', (code) => {
          if (code === 0) {
            if (verbose) {
              console.log('Build done successfully'.green, `${group}/${resource}`);
            }
            resolve();
            return;
          }
          if (verbose) {
            console.log('Build failed'.red, `${group}/${resource}`);
          }
          reject();
        });
      });
    });
  }

  buildFromQueue(group, resource) {
    delete this.buildQueue[`${group}/${resource}`];
    Builder.buildExercise(group, resource);
  }

  triggerBuildQueue(group, resource) {
    if (this.buildQueue[`${group}/${resource}`]) {
      return;
    }
    this.buildQueue[`${group}/${resource}`] = true;
    setTimeout(() => {
      this.buildFromQueue(group, resource);
    }, 150);
  }

  watch() {
    let ready = false;
    watcher.watch([`${srcPath}**/*.ts`, `${srcPath}*/*/*.html`], {
      ignoreInitial: true
    }).on('all', (event, path) => {
      path = path.replace(/\\/g, '/');
      console.log('changed', path);
      const resourceTest = path.match('src\\/([^\\_]\\w+)\\/(\\w+)\\/');
      if (!resourceTest) {
        console.log('Skipped, path not match'.gray);
        return;
      }
      const group = resourceTest[1];
      const resource = resourceTest[2];
      fs.stat(`${srcPath}${group}/${resource}`, (err, stat) => {
        if (err) {
          console.log('err');
          return;
        }
        if (!stat.isDirectory()) {
          console.log('notdir');
          return;
        }

        if (Path.extname(path) === '.html') {
          Builder.parseTypes(group, resource);
        }
        else {
          this.triggerBuildQueue(group, resource);
        }
      });
    });
  }

  static isGenerated(group, resource) {
    const resourcePath = `${srcPath}${group}/${resource}/`;
    try {
      fs.statSync(`${resourcePath}main.ts`);
      return true;
    }
    catch (e) {
      return false;
    }
  }

  static generate(group, resource, force = true, verbose = true) {
    if (!force) {
      if (Builder.isGenerated(group, resource)) {
        return;
      }
    }
    const resourcePath = `${srcPath}${group}/${resource}/`;
    const files = fs.readdirSync(blueprintPath);
    files.forEach((file) => {
      let content = fs.readFileSync(`${blueprintPath}${file}`, 'UTF-8');
      content = content.replace(/__group__/g, group);
      content = content.replace(/__name__/g, resource);
      let newFileName = file;
      if (file.indexOf('.def') >= 0) {
        newFileName = Path.basename(file, '.def');
      }
      fs.writeFileSync(`${resourcePath}${newFileName}`, content);
      if (verbose) {
        console.log(`Generated ${resourcePath}${newFileName}`);
      }
    });
    Builder.parseTypes(group, resource, verbose);
  }
}

module.exports = Builder;
